# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from enum import Enum
from azure.core import CaseInsensitiveEnumMeta


class DimensionScope(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The scope at which the quota is applied."""

    WORKSPACE = "Workspace"
    SUBSCRIPTION = "Subscription"


class ItemType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The type of item."""

    JOB = "Job"
    """A program, problem, or application submitted for processing."""
    SESSION = "Session"
    """A logical grouping of jobs."""


class JobStatus(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The status of the job."""

    WAITING = "Waiting"
    EXECUTING = "Executing"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    CANCELLED = "Cancelled"


class JobType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The type of job."""

    UNKNOWN = "Unknown"
    QUANTUM_COMPUTING = "QuantumComputing"
    OPTIMIZATION = "Optimization"


class JsonPatchOperation(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The operation to be performed."""

    ADD = "add"
    REMOVE = "remove"
    REPLACE = "replace"
    MOVE = "move"
    COPY = "copy"
    TEST = "test"


class MeterPeriod(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The time period in which the quota's underlying meter is accumulated. Based on calendar year.
    'None' is used for concurrent quotas.
    """

    NONE = "None"
    MONTHLY = "Monthly"


class ProviderAvailability(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Provider availability."""

    AVAILABLE = "Available"
    DEGRADED = "Degraded"
    UNAVAILABLE = "Unavailable"


class SessionJobFailurePolicy(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Policy controlling the behavior of the Session when a job in the session fails."""

    ABORT = "Abort"
    """New jobs submitted after a job fails will be rejected."""
    CONTINUE = "Continue"
    """New jobs submitted after a job fails will be accepted."""


class SessionStatus(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The status of the session."""

    WAITING = "Waiting"
    EXECUTING = "Executing"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    FAILURE_S_ = "Failure(s)"
    TIMED_OUT = "TimedOut"


class TargetAvailability(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Target availability."""

    AVAILABLE = "Available"
    DEGRADED = "Degraded"
    UNAVAILABLE = "Unavailable"
